#include "Matrix.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

/**
 * @brief Return true if matrix is not good
 *
 * @param mat The matrix to check
 * @return true Error in matrix detected
 * @return false Matrix is good
 */
bool _matrix_crash_test(const matrix *const mat)
{
    if (mat == NULL)
    {
        errno = EFAULT;
        return true;
    }
    if (mat->data == NULL)
    {
        errno = EFAULT;
        return true;
    }
    if (mat->col < 1 || mat->row < 1)
    {
        errno = EINVAL;
        return true;
    }
    return false;
}

bool matrix_is_equal(const matrix *const mat1, const matrix *const mat2)
{
    if (mat1 == NULL || mat2 == NULL)
    {
        errno = EFAULT;
        return false;
    }
    if (mat1->col != mat2->col || mat1->row != mat2->row)
        return false;

    for (uint32_t i = 0; i < mat1->col; i += 1)
        for (uint32_t j = 0; j < mat1->row; j += 1)
        {
            if (mat1->data[i][j] != mat2->data[i][j])
                return false;
        }
    return true;
}

/**
 * @brief Create matrix and return NULL with errno set if error
 *
 * @param col number of col
 * @param row number of row
 * @return matrix* The pointer of new matrix of NULL for error
 */
matrix *matrix_create(uint32_t col, uint32_t row)
{
    if (col < 1 || row < 1)
    {
        errno = EINVAL;
        return NULL;
    }
    matrix *m = calloc(1, sizeof(matrix));
    m->col = col;
    m->row = row;
    m->data = calloc(col, sizeof(T *));
    for (uint32_t i = 0; i < col; i += 1)
    {
        m->data[i] = calloc(row, sizeof(T));
    }
    return m;
}

T matrix_max(matrix *orig)
{
    T res = 0;
    for (uint32_t i = 2; i < orig->col - 2; i += 1)
        for (uint32_t j = 2; j < orig->row - 2; j += 1)
        {
            // printf("Value : %d\n",orig->data[i][j]);
            if (orig->data[i][j] > res)
                res = orig->data[i][j];
        }
    return res;
}

T matrix_min(matrix *orig)
{
    T res = UINT16_MAX;
    for (uint32_t i = 2; i < orig->col - 2; i += 1)
        for (uint32_t j = 2; j < orig->row - 2; j += 1)
        {
            if (orig->data[i][j] < res)
                res = orig->data[i][j];
        }
    return res;
}

void matrix_init(matrix *mat, uint32_t col, uint32_t row)
{
    mat->col = col;
    mat->row = row;
    mat->data = calloc(col, sizeof(T *));
    for (uint32_t i = 0; i < col; i += 1)
    {
        mat->data[i] = calloc(row, sizeof(T));
    }
}

/**
 * @brief Free all data in matrix* alloced with matrix_create
 *
 * @param mat The pointer of matrix to free
 */
void matrix_destroy(matrix *mat)
{
    if (_matrix_crash_test(mat))
        return;

    for (uint32_t i = 0; i < mat->col; i += 1)
        free(mat->data[i]);
    free(mat->data);
    free(mat);
}

/**
 * @brief Free matrix create in stack but passed in param
 *
 * @param mat
 */
void matrix_free(matrix *mat)
{
    if (_matrix_crash_test(mat))
        return;

    for (uint32_t i = 0; i < mat->col; i += 1)
        free(mat->data[i]);
    free(mat->data);
}

/**
 * @brief Functiob for print all data from matrix
 *
 * @param mat The matrix to print
 */
void matrix_print(const matrix *const mat)
{
    if (_matrix_crash_test(mat))
        return;

    printf("\n");
    for (uint32_t i = 0; i < mat->col; i += 1)
    {
        printf("%d\t", i);
    }
    printf("Colunm = %u\n", mat->col);

    for (uint32_t i = 0; i < mat->col * 10; i += 1)
    {
        printf("─");
    }
    printf("\n");

    for (uint32_t i = 0; i < mat->row; i += 1)
    {
        for (uint32_t j = 0; j < mat->col; j += 1)
        {
            printf("%d\t", mat->data[j][i]);
        }
        printf("  | %d", i);
        printf("\n");
    }
    for (uint32_t i = 0; i < mat->col; i += 1)
        printf("\t");
    printf("Line = %u\n", mat->row);
    printf("\n");
}

/**
 * @brief Set all data in matrix to values return errno if error
 *
 * @param mat
 * @param value
 */
void matrix_set_values(matrix *mat, T value)
{
    if (_matrix_crash_test(mat))
        return;

    for (uint32_t i = 0; i < mat->col; i += 1)
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            mat->data[i][j] = value;
        }
}

void _print_data_table(uint32_t data_length, T *data)
{
    for (uint32_t i = 0; i < data_length; i += 1)
        printf("%d-", data[i]);
}

/**
 * @brief Create matrix from array passed in param mat = NULL, errno is set for error
 *
 * @param mat The double pointer for matrix create
 * @param col The number of column
 * @param row The number of row
 * @param data The array of data to use for set matrix
 * @param data_length The number of data in array
 */
matrix *matrix_create_from_array(uint32_t col, uint32_t row, T *data, uint32_t data_length)
{
    if (col == 0 || row == 0)
    {
        errno = EINVAL;
        return NULL;
    }
    if (data == NULL || data_length < 1 || (col * row) > data_length)
    {
        errno = EFAULT;
        return NULL;
    }

    matrix *res = matrix_create(col, row);
    for (uint32_t i = 0, k = 0; i < res->col; i += 1)
    {
        for (uint32_t j = 0; j < res->row; j += 1, k += 1)
        {
            res->data[i][j] = data[k];
        }
    }
    return res;
}

/**
 * @brief Convert matrix to array 1D malloced errno is set for error
 *
 * @param mat The matrix to get all data
 * @param table The table to set
 */
void matrix_to_array(const matrix *const mat, T *table)
{
    if (_matrix_crash_test(mat))
        return;
    if (table == NULL)
    {
        errno = EINVAL;
        return;
    }

    for (uint32_t i = 0; i < mat->col; i += 1)
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            table[i + j] = mat->data[i][j];
        }
}

/**
 * @brief Return the cloned matrix return NULL for error with errno set
 *
 * @param mat The matrix to use for clone
 * @return matrix* Return the pointer of new matrix clone from mat
 */
matrix *matrix_clone(const matrix *const mat)
{
    if (_matrix_crash_test(mat))
        return NULL;
    if (mat->col < 1 || mat->row < 1)
        return NULL;

    matrix *res = matrix_create(mat->col, mat->row);
    for (uint32_t i = 0; i < mat->col; i += 1)
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            res->data[i][j] = mat->data[i][j];
        }
    return res;
}

/**
 * @brief Get data from matrix at index x,y errno is set for error
 *
 * @param elem The pointer of element with value get
 * @param mat The matrix original
 * @param id_x The index x
 * @param id_y The index y
 */
void matrix_get(T *elem, const matrix *const mat, uint32_t id_x, uint32_t id_y)
{
    if (_matrix_crash_test(mat))
        return;
    if (id_x > 0 && id_x < mat->col && id_y > 0 && id_y < mat->row)
        *elem = mat->data[id_x][id_y];
    else
        errno = EINVAL;
}

/**
 * @brief Set data for matrix at index x,y errno is set for error
 *
 * @param mat The matrix to set
 * @param id_x The index x
 * @param id_y The index x
 * @param elem The elem to use for set matrix
 */
void matrix_set(const matrix *mat, uint32_t id_x, uint32_t id_y, const T elem)
{
    if (_matrix_crash_test(mat))
        return;
    if (id_x > 0 && id_x < mat->col && id_y > 0 && id_y < mat->row)
        mat->data[id_x][id_y] = elem;
    else
        errno = EINVAL;
}

/**
 * @brief Apply func to all data in matrix
 *
 * @param mat The matrix to use
 * @param func The function to use for each data in matrix
 */
void matrix_map_ip(const matrix *mat, void (*func)(T *))
{
    if (_matrix_crash_test(mat))
        return;
    for (uint32_t i = 0; i < mat->col; i += 1)
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            func(&mat->data[i][j]);
        }
}

/**
 * @brief Apply function for all element in matrix and return the new matrix return NULL for error with errno set
 *
 * @param mat The matrix source to use
 * @param func The function to use for all data in matrix mat
 * @return matrix* Return the new matrix create with all data from mat with function apply
 */
matrix *matrix_map(const matrix *const mat, T (*func)(T *))
{
    if (_matrix_crash_test(mat))
        return NULL;
    matrix *res = matrix_create(mat->col, mat->row);
    for (uint32_t i = 0; i < mat->col; i += 1)
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            res->data[i][j] = func(&mat->data[i][j]);
        }
    return res;
}

/**
 * @brief Create matrix with padding and return it return NULL for error with errno set
 *
 * @param mat The original matrix to use for padding
 * @param number_of_padding The number of padding for the new matrix
 * @return matrix* The new matrix with padding added from mat
 */
matrix *matrix_padding(const matrix *const mat, uint32_t number_of_padding)
{
    if (_matrix_crash_test(mat))
        return NULL;
    uint32_t size_x = mat->col;
    uint32_t size_y = mat->row;
    matrix *res = matrix_create(size_x + (2 * number_of_padding), size_y + (2 * number_of_padding));

    for (uint32_t i = 0; i < mat->col; i += 1)
    {
        for (uint32_t j = 0; j < mat->row; j += 1)
        {
            res->data[i + number_of_padding][j + number_of_padding] = mat->data[i][j];
        }
    }
    return res;
}

/**
 * @brief Create submatrix from orig matrix and return the new matrix alloced
 *
 * @param mat The original matrix to use
 * @param x0 The x0 position in original matrix
 * @param x1 The x1 position in original matrix
 * @param y0 The y0 position in original matrix
 * @param y1 The y1 position in original matrix
 * @return matrix* The new submatrix from mat
 */
matrix *matrix_extract_submatrix(const matrix *const mat, uint32_t x0, uint32_t x1, uint32_t y0, uint32_t y1)
{
    if (_matrix_crash_test(mat))
        return NULL;

    int32_t row = x1 - x0;
    int32_t col = y1 - y0;

    matrix *res = matrix_create(col, row);

    for (int i = 0; i < col; i += 1)
    {
        for (int j = 0; j < row; j += 1)
        {
            res->data[i][j] = mat->data[x0 + i][y0 + j];
        }
    }
    return res;
}

/**
 * @brief Return the transposed matrix from the original matrix
 *
 * @param mat The original matrix to use
 * @return matrix* The new transposed matrix
 */
matrix *matrix_transpose(const matrix *const mat)
{
    if (_matrix_crash_test(mat))
        return NULL;
    matrix *res = matrix_create(mat->row, mat->col);

    for (uint32_t i = 0, k = 0; i < mat->col; i += 1, k = 0)
    {
        for (uint32_t j = 0; j < mat->row; j += 1, k += 1)
        {
            res->data[k][i] = mat->data[i][j];
        }
    }

    return res;
}